# pylint: disable=missing-module-docstring, missing-function-docstring
import pandas as pd
import pytest

from xil._currencies import CurrencyCode
from xil.mercantile import get_mercantile_df

from .test_discount import test_df as _test_df_helper


@pytest.fixture(name="df")
def df_fixture() -> pd.DataFrame:
    return get_mercantile_df()


@pytest.fixture(name="currencies")
def currencies_fixture() -> set[CurrencyCode]:
    return {
        CurrencyCode.CAD,
        CurrencyCode.MXN,
        CurrencyCode.EGP,
        CurrencyCode.RUB,
        CurrencyCode.JPY,
        CurrencyCode.LBP,
        CurrencyCode.XAG,
        CurrencyCode.XAU,
        CurrencyCode.CNY,
        CurrencyCode.SEK,
        CurrencyCode.ZAR,
        CurrencyCode.SAL,
        CurrencyCode.USD,
        CurrencyCode.XPT,
        CurrencyCode.SGD,
        CurrencyCode.INR,
        CurrencyCode.AUD,
        CurrencyCode.BRL,
        CurrencyCode.CHF,
        CurrencyCode.JOD,
        CurrencyCode.NOK,
        CurrencyCode.PLN,
        CurrencyCode.DKK,
        CurrencyCode.GBP,
        CurrencyCode.XPD,
        CurrencyCode.NZD,
        CurrencyCode.HKD,
        CurrencyCode.HUF,
        CurrencyCode.TRY,
        CurrencyCode.EUR,
    }


@pytest.mark.parametrize("dropped_currencies", [[CurrencyCode.XAU]])
@pytest.mark.live
def test_df(
    df: pd.DataFrame,
    currencies: set[CurrencyCode],
    dropped_currencies: list[CurrencyCode],
) -> None:
    _test_df_helper(df, currencies, dropped_currencies=dropped_currencies)



# pylint: disable=missing-function-docstring
"""
Disable network option and marker for pytest.

It can be used as a flag, for example:
```sh
pytest . -m "not live" --disable-network
```

Or as a marker (for debugging purposes):
```python
@pytest.mark.disable_network
def test_something():
    ...
```

Based on:
https://github.com/best-doctor/pytest_network
"""

import socket
from collections.abc import Iterator
from typing import NoReturn

import pytest
from pytest import Config, FixtureRequest, Item, Parser

_original_connect = socket.socket.connect
_DISABLE_NETWORK = "disable_network"


class _NetworkUsageError(Exception):
    pass


def _patched_connect(self, address) -> NoReturn:  # type: ignore[no-untyped-def]
    raise _NetworkUsageError(
        "The test attempts to connect to the network. If it is expected and required "
        "for the test, enable it with the `live` test marker: `@pytest.mark.live`."
    )


@pytest.fixture
def disable_network() -> Iterator[None]:
    socket.socket.connect = _patched_connect  # type: ignore[method-assign]
    yield
    socket.socket.connect = _original_connect  # type: ignore[method-assign]


def pytest_addoption(parser: Parser) -> None:
    group = parser.getgroup("network")
    group.addoption(
        "--disable-network",
        action="store_true",
        dest=_DISABLE_NETWORK,
        help="Monkeypatch socket.socket.connect to disable network connection.",
    )


@pytest.fixture(autouse=True)
def _network_marker(request: FixtureRequest) -> None:
    if request.config.getoption(_DISABLE_NETWORK):
        request.getfixturevalue("disable_network")


def pytest_configure(config: Config) -> None:
    config.addinivalue_line(
        "markers", f"{_DISABLE_NETWORK}: disable network in marked tests."
    )


def pytest_runtest_setup(item: Item) -> None:
    if (
        list(item.iter_markers(name=_DISABLE_NETWORK))
        and socket.socket.connect != _patched_connect
    ):
        socket.socket.connect = _patched_connect  # type: ignore[method-assign]


def pytest_runtest_teardown(item: Item) -> None:
    if (
        list(item.iter_markers(name=_DISABLE_NETWORK))
        and socket.socket.connect != _original_connect
    ):
        socket.socket.connect = _original_connect  # type: ignore[method-assign]



# pylint: disable=missing-module-docstring, missing-function-docstring
import pandas as pd
import pytest

from xil._currencies import CurrencyCode
from xil.discount import get_discount_df


@pytest.fixture(name="df")
def df_fixture() -> pd.DataFrame:
    return get_discount_df()


@pytest.fixture(name="currencies")
def currencies_fixture() -> set[CurrencyCode]:
    return {
        CurrencyCode.EUR,
        CurrencyCode.ETB,
        CurrencyCode.LKR,
        CurrencyCode.NGN,
        CurrencyCode.KRW,
        CurrencyCode.BGN,
        CurrencyCode.AUD,
        CurrencyCode.CHF,
        CurrencyCode.CNY,
        CurrencyCode.GBP,
        CurrencyCode.MXN,
        CurrencyCode.EGP,
        CurrencyCode.SAR,
        CurrencyCode.HUF,
        CurrencyCode.CAD,
        CurrencyCode.HKD,
        CurrencyCode.INR,
        CurrencyCode.PEN,
        CurrencyCode.USD,
        CurrencyCode.XPD,
        CurrencyCode.IDR,
        CurrencyCode.RUB,
        CurrencyCode.TRY,
        CurrencyCode.XPT,
        CurrencyCode.SEK,
        CurrencyCode.TWD,
        CurrencyCode.HRK,
        CurrencyCode.NOK,
        CurrencyCode.SAL,
        CurrencyCode.SGD,
        CurrencyCode.CZK,
        CurrencyCode.PHP,
        CurrencyCode.NZD,
        CurrencyCode.XAU,
        CurrencyCode.LBP,
        CurrencyCode.PLN,
        CurrencyCode.XAG,
        CurrencyCode.BRL,
        CurrencyCode.CLP,
        CurrencyCode.JPY,
        CurrencyCode.ZAR,
        CurrencyCode.ARS,
        CurrencyCode.DKK,
        CurrencyCode.JOD,
        CurrencyCode.RON,
    }


@pytest.mark.parametrize(
    "dropped_currencies", [[CurrencyCode.XAU, CurrencyCode.NGN]]
)  # There's an issue with Gold
@pytest.mark.live
def test_df(
    df: pd.DataFrame,
    currencies: set[CurrencyCode],
    dropped_currencies: list[CurrencyCode],
) -> None:
    assert set(df.index) == currencies
    df = df.drop(index=dropped_currencies)
    for method in ("cash", "transfer"):
        assert (df[(method, "sell")] >= df[(method, "buy")]).all()



"""
Test the _currencies module.
This test module checks only a small subset of the currencies.
"""

# pylint: disable=missing-function-docstring,missing-class-docstring
import pytest

from xil._currencies import (
    CurrencyCode,
    CurrencyNotSupportedError,
    currency_code_from_heb_name,
    currency_from_heb_name,
    optional_currency_from_heb_name,
)


def test_currency_code_from_heb_name() -> None:
    assert currency_code_from_heb_name('דולר ארה"ב') == CurrencyCode.USD


def test_currency_from_heb_name() -> None:
    assert currency_from_heb_name("פרנק שויצרי") == "CHF"


@pytest.mark.parametrize("currency_name", ["מטבע לא קיים", "NOT A CURRENCY"])
class TestNonExistingCurrencies:
    @staticmethod
    def test_not_supported_currency(currency_name: str) -> None:
        with pytest.raises(
            CurrencyNotSupportedError, match=f"Unknown currency name: '{currency_name}'"
        ):
            currency_code_from_heb_name(currency_name)

    @staticmethod
    def test_optional_currency_returns_none(currency_name: str) -> None:
        assert optional_currency_from_heb_name(currency_name) is None



"""
Test the _df_normalizer module.
"""

# pylint: disable=missing-function-docstring,missing-class-docstring
from contextlib import AbstractContextManager
from unittest.mock import Mock, patch

import pandas as pd
import pytest

from xil._df_normalizer import (
    _CURRENCY_CODE_KEY,
    _CURRENCY_KEY,
    _CURRENCY_NAME_KEY,
    BaseDataFrameNormalizer,
    DataFrameNormalizer,
    JPYNormalizer,
)


@pytest.fixture(name="currencies_df")
def fixture_currencies_df() -> pd.DataFrame:
    return pd.DataFrame(
        [['דולר ארה"ב', 1, 3.41], ["ין יפני", 100, 2.65]],
        columns=[_CURRENCY_NAME_KEY, (_CURRENCY_KEY, "amount"), ("transfer", "sell")],
    )


@pytest.fixture(name="currencies_df_with_code")
def fixture_currencies_df_with_code(currencies_df: pd.DataFrame) -> pd.DataFrame:
    DataFrameNormalizer(currencies_df).add_code_from_name()
    return currencies_df


class TestBaseNormalizer:
    @staticmethod
    def test_set_index(currencies_df_with_code: pd.DataFrame) -> None:
        BaseDataFrameNormalizer(currencies_df_with_code).set_code_index()
        assert currencies_df_with_code.index.name == _CURRENCY_CODE_KEY

    @staticmethod
    def test_norm(currencies_df: pd.DataFrame) -> None:
        df_normalizer = BaseDataFrameNormalizer(currencies_df)
        with patch.object(
            BaseDataFrameNormalizer, "set_code_index", autospec=True
        ) as mock:
            df_normalizer.norm()
            mock.assert_called_once_with(df_normalizer)


class TestNormalizer:
    @staticmethod
    def _patch_normalizer_attr(attr: str) -> AbstractContextManager[Mock]:
        return patch.object(DataFrameNormalizer, attr, autospec=True)

    @staticmethod
    def test_drop_name(currencies_df: pd.DataFrame) -> None:
        assert _CURRENCY_NAME_KEY in currencies_df.columns
        DataFrameNormalizer(currencies_df).drop_currency_name()
        assert _CURRENCY_NAME_KEY not in currencies_df.columns

    @staticmethod
    def test_add_code(currencies_df: pd.DataFrame) -> None:
        assert _CURRENCY_CODE_KEY not in currencies_df.columns
        DataFrameNormalizer(currencies_df).add_code_from_name()
        assert _CURRENCY_CODE_KEY in currencies_df.columns

    @staticmethod
    def test_preprocess_names(currencies_df: pd.DataFrame) -> None:
        names = currencies_df[_CURRENCY_NAME_KEY]
        with patch.object(
            DataFrameNormalizer, "preprocess_names", autospec=True
        ) as mock:
            DataFrameNormalizer(currencies_df).add_code_from_name()
            mock.assert_called_once_with(names)

    @classmethod
    def test_norm(cls, currencies_df: pd.DataFrame) -> None:
        df_normalizer = DataFrameNormalizer(currencies_df)
        with (
            cls._patch_normalizer_attr("add_code_from_name") as mock_add_code,
            cls._patch_normalizer_attr("set_code_index") as mock_set_index,
            cls._patch_normalizer_attr("drop_currency_name") as mock_drop_name,
        ):
            df_normalizer.norm()
            for mock in [mock_add_code, mock_set_index, mock_drop_name]:
                mock.assert_called_once_with(df_normalizer)


class TestJPYNormalizer:
    @staticmethod
    @pytest.fixture(name="jpy_df")
    def fixture_jpy_df() -> pd.DataFrame:
        return pd.DataFrame(
            [['דולר ארה"ב', 3.41], ["100 ין יפני", 2.65]],
            columns=[_CURRENCY_NAME_KEY, ("transfer", "sell")],
        )

    @staticmethod
    @pytest.fixture(name="normed_jpy_df")
    def fixture_normed_jpy_df() -> pd.DataFrame:
        return pd.DataFrame(
            [[3.41], [2.65]],
            columns=[("transfer", "sell")],
            index=pd.Index(["USD", "JPY"], name=_CURRENCY_CODE_KEY),
        )

    @staticmethod
    @pytest.mark.parametrize(
        ("raw_name", "normed_name"),
        [
            ("100 ין יפני", "ין יפני"),
            ("ין יפני 100", "ין יפני"),
            (" ין יפני 100 ", "ין יפני 100"),
            ("דולר", "דולר"),
            ("1 דולר", "1 דולר"),
            (" 100 דולר", "100 דולר"),
            ("100 דולר", "דולר"),
        ],
    )
    def test_remove_100(raw_name: str, normed_name: str) -> None:
        # pylint: disable=protected-access
        assert JPYNormalizer._remove_100(raw_name) == normed_name

    @staticmethod
    def test_norm(jpy_df: pd.DataFrame, normed_jpy_df: pd.DataFrame) -> None:
        assert JPYNormalizer(jpy_df).norm().equals(normed_jpy_df), (
            "The normalized `jpy_df` is different than expected"
        )



"""
Test the known list of banks against the Bank of Israel (BOI) online list
"""

import pandas as pd
import pytest

from xil.boi import get_boi_url_response

BOI_XML_URL = "\
https://www.boi.org.il/boi_files/Pikuah/banking_corporations_en.xml"

KNOWN_BANKS = {
    "Bank Hapoalim B.M",
    "Bank Leumi Le-Israel B.M",
    "Bank Massad Ltd",
    "Bank of Jerusalem Ltd",
    "Israel Discount Bank Ltd",
    "Mercantile Discount Bank ltd",
    "Mizrahi Tefahot Bank Ltd",
    "The First International Bank of Israel Ltd",
    "Bank Yahav  for Government Employees Ltd",  # Not supported - no public data
    "One Zero Digital Bank LTD",  # Not supported - no public data
    "Bank Esh Israel Ltd",  # Not supported - not commercially active yet
}


@pytest.fixture(name="boi_banks")
def fixture_boi_banks() -> set[str]:
    """Get the set on Israeli banks with bank code from BOI online XML"""
    with get_boi_url_response(BOI_XML_URL) as response:
        df = pd.read_xml(response)
    return set(
        df[(df["Category"] == "COMMERCIAL BANKS") & df["Bank_Code"].notna()]["Name"]
    )


@pytest.mark.live
def test_boi_banks(boi_banks: set[str]) -> None:
    """Test the online set vs. the hard-coded one"""
    assert boi_banks == KNOWN_BANKS, (
        "Mismatch between the updated banks list data from BOI and the saved data"
    )



# pylint: disable=missing-module-docstring
import pandas as pd

from xil._currencies import CurrencyCode


def _test_df_sanity(df: pd.DataFrame, expected_currencies: list[CurrencyCode]) -> None:
    """Test the actual vs. expected currencies and rates"""
    assert (df.index == expected_currencies).all(), "The currencies are not as expected"
    assert (df[("transfer", "sell")] > df[("transfer", "buy")]).all()
    assert (df[("cash", "sell")] > df[("cash", "buy")]).all()
    assert (df[("cash", "sell")] > df[("transfer", "sell")]).all()
    assert (df[("transfer", "buy")] > df[("cash", "buy")]).all()



"""
Test the _headers module.
"""

import ssl
from collections.abc import Generator
from unittest.mock import Mock, patch
from urllib.request import Request

import pytest

from xil._headers import _DEFAULT_CONTEXT, get_url_response


@pytest.fixture(name="url")
def fixture_url() -> str:
    """A tests URL that can be used (no actual call is made)"""
    return "http://httpbin.org/get"


@pytest.fixture(name="mock_urlopen")
def fixture_mock_urlopen() -> Generator[Mock, None, None]:
    """Mock the call to an external URL"""
    with patch("urllib.request.urlopen", autospec=True) as mock:
        yield mock


def _compare_requests(request1: Request, request2: Request) -> bool:
    """
    This is required since two Request objects with the same data are not equal.
    Compare here only the attributes of interest.
    """
    return (
        request1.full_url == request2.full_url and request1.headers == request2.headers
    )


@pytest.mark.parametrize(
    ("set_context", "expected_context"), [(False, None), (True, _DEFAULT_CONTEXT)]
)
def test_get_url_response_context(
    url: str,
    set_context: bool,
    expected_context: ssl.SSLContext | None,
    mock_urlopen: Mock,
) -> None:
    """Test that get_url_response sets an SSL context when it is asked to"""
    get_url_response(url, set_context=set_context)
    mock_urlopen.assert_called_once()
    assert mock_urlopen.call_args.kwargs == {
        "context": expected_context,
    }, "The context passed to urlopen is different than expected"



# pylint: disable=missing-module-docstring, missing-function-docstring, redefined-outer-name
import pandas as pd
import pytest

from xil._currencies import CurrencyCode
from xil.leumi import get_leumi_df

from ._test_df import _test_df_sanity


@pytest.fixture
def df() -> pd.DataFrame:
    return get_leumi_df()


@pytest.fixture()
def expected_currencies() -> list[CurrencyCode]:
    return [
        CurrencyCode.USD,
        CurrencyCode.EUR,
        CurrencyCode.GBP,
        CurrencyCode.JPY,
        CurrencyCode.CHF,
        CurrencyCode.AUD,
        CurrencyCode.CAD,
        CurrencyCode.ZAR,
        CurrencyCode.AED,
        CurrencyCode.SEK,
        CurrencyCode.NOK,
        CurrencyCode.DKK,
    ]


@pytest.mark.live
def test_df(df: pd.DataFrame, expected_currencies: list[CurrencyCode]) -> None:
    _test_df_sanity(df, expected_currencies)



# pylint: disable=missing-module-docstring, missing-function-docstring, redefined-outer-name
import calendar
import os
from datetime import date

import pandas as pd
import pytest

from xil._currencies import CurrencyCode
from xil.poalim import _get_url, get_df

from ._test_df import _test_df_sanity

skip_in_github_actions = pytest.mark.skipif(
    condition=os.getenv("CI") == "true",
    reason="Poalim API is requires human verification outside Israel",
)


def _is_weekend(t: date) -> bool:
    """Return True if the day is Saturday or Sunday, otherwise False"""
    return t.weekday() in (calendar.SATURDAY, calendar.SUNDAY)


@pytest.mark.parametrize(
    ("t", "expected_url"),
    [
        (
            date(2345, 10, 30),
            "https://www.bankhapoalim.co.il/he/coin-rates?date=2345-10-30",
        ),
        (
            date(2020, 4, 8),
            "https://www.bankhapoalim.co.il/he/coin-rates?date=2020-04-08",
        ),
    ],
)
def test_get_url(t: date, expected_url: str) -> None:
    assert _get_url(t) == expected_url


@pytest.mark.live
@skip_in_github_actions
@pytest.mark.parametrize(
    "t",
    [
        pytest.param(date(2024, 2, 3), id="Saturday 2023-02-03"),
        pytest.param(date(2024, 2, 4), id="Sunday 2023-02-04"),
    ],
)
def test_weekend_emptiness(t: date) -> None:
    """Weekend in this context is Saturday and Sunday"""
    assert _is_weekend(t)
    assert get_df(t).empty, "The df is nonempty"


@pytest.fixture()
def df() -> pd.DataFrame:
    return get_df()


@pytest.fixture()
def multi_date_df() -> pd.DataFrame:
    return get_df(keep_last_date_only=False)


@pytest.fixture()
def expected_currencies() -> list[CurrencyCode]:
    return [
        CurrencyCode.USD,
        CurrencyCode.GBP,
        CurrencyCode.CHF,
        CurrencyCode.DKK,
        CurrencyCode.EUR,
        CurrencyCode.NOK,
        CurrencyCode.CAD,
        CurrencyCode.AUD,
        CurrencyCode.JPY,
        CurrencyCode.JOD,
        CurrencyCode.TRY,
    ]


@pytest.mark.live
@skip_in_github_actions
def test_df(df: pd.DataFrame, expected_currencies: list[CurrencyCode]) -> None:
    _test_df_sanity(df, expected_currencies)


@pytest.mark.live
@skip_in_github_actions
def test_multi_date_df(multi_date_df: pd.DataFrame) -> None:
    assert len(multi_date_df.date.unique()) > 1, "Multiple dates are expected"



# pylint: disable=missing-module-docstring, missing-function-docstring
import pandas as pd
import pytest

from xil._currencies import CurrencyCode
from xil.onezero import get_onezero_df


@pytest.fixture(name="df")
def df_fixture() -> pd.DataFrame:
    return get_onezero_df()


@pytest.mark.live
@pytest.mark.skip(
    reason="One-zero data is not available on https://www.onezerobank.com/currencies/"
)
def test_df(df: pd.DataFrame) -> None:
    assert (df.index == [CurrencyCode.EUR, CurrencyCode.USD]).all(), (
        "The currencies are not as expected"
    )
    assert (df[("transfer", "buy")] < df[("transfer", "sell")]).all(), (
        "The buy rate is not lower than the sell rate"
    )
